/***************************************************************************
 *
 * linux_lkm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * This file contains the code needed to load, initialize, and unload the
 * lomac Loadable Kernel Module (LKM).
 *
 * References:
 *
 * David A. Rusling `The Linux Kernel' chapter 12: Modules, March 1998.
 * (A description of the linux kernel, available on the web.)
 *
 * Allesandro Rubini `Dynamic Kernels: Modularized Device Drivers' in
 * Linux Journal #23, March 1996.
 *
 * Allesandro Rubini `Dynamic Kernels: Discovery' in Linux Journal #24,
 * April 1996.
 *
 * Michael K. Johnson `The Linux Kernel Hacker's Guide' (a moderated communal
 * web document). 
 *
 ***************************************************************************/

#include <linux/module.h>      /* for module macros */
#include <linux/kernel.h>      /* for kernel utility functions */
#include <linux/types.h>       /* for gid_t */
#include <linux/netdevice.h>   /* for struct device    */

#include "kernel_lpid.h"
#include "kernel_lgid.h"
#include "kernel_lvnode.h"
#include "kernel_lnic.h"
#include "kernel_driver.h"
#include "kernel_syscalls.h"
#include "kernel_binfmt.h"

#include "lomac_nlm.h"
#include "lomac_tpl.h"
#include "lomac_plm.h"
#include "lomac_fsolm.h"


/* This string puts the version number in the binary. */
static char *banner_s = 
"LOMAC v0.2 for Linux 2.2 Copyright (C) 1999 NAI Labs.\n"
"LOMAC is free software covered by the GNU General Public License version 2.\n"
"There is absolutely no warranty for LOMAC.  See the LICENSE file for details.";


/* init_module()
 *
 * in:     nothing
 * out:    nothing
 * return: value      condition
 *         -----      ---------
 *         0          success
 *
 * This function is called by the kernel when the lomac lkm is loaded.  It
 * initializes all of the lomac data structures, and patches the kernel's
 * system call vector to allow the lomac lkm to examine calls and perform
 * access control mediation on them.
 */

int init_module( void ) {

  /* Set up interception of the system calls which are interesting to Lomac. */
  REPLACE_SYSCALL( read,         lomac_read   );
  REPLACE_SYSCALL( readv,        lomac_readv  );
  REPLACE_SYSCALL( fchown,       lomac_fchown );
  REPLACE_SYSCALL( chown,        lomac_chown  );
  REPLACE_SYSCALL( open,         lomac_open   );
  REPLACE_SYSCALL( pipe,         lomac_pipe   );
  REPLACE_SYSCALL( socketcall,   lomac_socketcall );
  REPLACE_SYSCALL( kill,         lomac_kill   );
  REPLACE_SYSCALL( write,        lomac_write  );
  REPLACE_SYSCALL( writev,       lomac_writev );
  REPLACE_SYSCALL( unlink,       lomac_unlink );
  REPLACE_SYSCALL( rmdir,        lomac_rmdir  );
  REPLACE_SYSCALL( truncate,     lomac_truncate );
  REPLACE_SYSCALL( ftruncate,    lomac_ftruncate );
  REPLACE_SYSCALL( setpgid,      lomac_setpgid );
  REPLACE_SYSCALL( setsid,       lomac_setsid );


  register_lomac_binfmt();

  /* Initialize Lomac data structures */
  init_tpl();
  init_plm();
  init_fsolm();
  init_nlm();
  init_driver();


  /* Print a victory message to the console, indicating a successful load. */
  printk( "loaded %s\n", banner_s );
  return( 0 );

} /* init_module() */


/* cleanup_module()
 * 
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 * This function is called by the kernel in response to a request to unload
 * the lomac lkm.  Someday, this function should unpatch the system call
 * vector so the kernel can keep running once the LKM is gone, but this
 * requires some method to rescue processes that are currently blocked
 * in lomac-intercepted system calls.
 */

void cleanup_module( void ) {

  /* Remove interceptions; restore `sys_call_table' to it's original state. */
  /* RESTORE_SYSCALL( ... ); */

  printk( "unloaded %s\n", banner_s );

} /* cleanup_module() */


