/***************************************************************************
 *
 * linux_syscalls.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * This file contains the functions used by the lomac lkm to replace 
 * some of the kernel's system call functions.  The lomac lkm "intercepts"
 * a strategic subset of the kernel's system calls in order to keep track
 * of what processes exist and what they are doing, and to perform mediation.
 * For each intercepted system call function sys_FOO(), there is a lomac-
 * specifc function lomac_FOO() in this file.  Typically, each lomac_FOO()
 * function has three sequential parts:
 * 1. Examine the parameters specified by the calling process, and either
 *    update some lomac-specific data structure or perform some access
 *    control.
 * 2. Call the kernel's original sys_FOO() function.
 * 3. Possibly update some lomac-specific data structure based on the
 *    result of the original kernel function.
 *
 * Note that in step 2, the lomac_FOO() function calls the sys_FOO()
 * function through the `orig_FOO' pointer rather than by name.  Either
 * way would work, but calling through the pointer allows a given system
 * call to be intercepted by multiple modules.  Provided that every
 * module performs the call through an `orig' pointer as this one does,
 * each module's function will be called in the reverse of the order
 * in which they modified the system call table, until the original
 * kernel function is called at last.
 *
 * Note: all of the socket-related system calls that are separate under
 *       BSD are collected under a single sys_socketcall system call in
 *       Linux.  See net/socket.c .
 *
 * References:
 *
 * McKusick et. al. `The Design and Implementation of the 4.4 BSD
 * Operating System,' Addison-Wesley 1996.  Chapter 11: Interprocess
 * Communication has good info on sockets relevant to
 * lomac_socketcall().  Section 4.8 explains job control, a topic
 * relevant to lomac_setpgid() and lomac_setsid().
 *
 ***************************************************************************/

#include <linux/kernel.h>
#include <linux/sched.h>   /* for `current' process */
#include <linux/fs.h>      /* for struct inode needed by lomac_monitor.h */
#include <linux/net.h>     /* for SYS_BIND */
#include <net/sock.h>      /* for struct sock, struct unix_opt */
#include <asm/uaccess.h>   /* for copy_from_user() */

#include "kernel_lgid.h"           /* for lgid_t   */
#include "kernel_lpid.h"           /* for lpid_t   */
#include "kernel_lvnode.h"         /* for lvnode_t */
#include "kernel_lprocess.h"       /* for lprocess_from_pid() */

#include "lomac_level.h"   /* for level_t, needed by lomac_monitor.h */
#include "kernel_syscalls.h"
#include "lomac_monitor.h"
#include "lomac_plm.h"
#include "lomac_fsolm.h"
#include "lomac_tpl.h"


/* --------------------------- Private Functions ------------------------- */

struct inode *fd_to_inode( unsigned long );


/* fd_to_inode()
 *
 * in:     fd
 * out:    nothing
 * return: pointer to inode corresponding to `fd' in the `current' process
 *         or NULL if there is no such inode.
 *
 */

struct inode *fd_to_inode( unsigned long fd ) {

  struct file *p_file;      /* pointer to file struct for `fd' */

 /* Make `p_file' point to the file corresponding to `fd'.  Like *
  * sys_fstat(), we're not going to use fget().  We just want a  *
  * brief atomic peek at the inode, so we're not going to fool   *
  * with all that reference-counting fget() and fput() jazz.     */
  if( fd >= NR_OPEN || !( p_file = current->files->fd[ fd ] ) ) {
    return( NULL );
  }

  /* return inode pointer */
  return( p_file->f_dentry->d_inode );

} /* fd_to_inode() */




/* --------------------------- Public Functions ------------------------- */


int lomac_read( unsigned int fd, char *buf, int count ) {

  struct inode *p_inode;              /* inode corresponding to `fd' */
  int ret_val;                        /* holds return value */

  if( ( p_inode = fd_to_inode( fd ) ) ) {
    monitor_read( p_inode, "read" );
  }

  ret_val = ((int (*)(unsigned int, char *, int))orig_read)( fd, buf, count );

  return( ret_val );
} /* lomac_read() */


int lomac_readv( unsigned long fd, const struct iovec *vector, long count ) {

  struct inode *p_inode;              /* inode corresponding to `fd' */
  int ret_val;                        /* holds return value */

  if( ( p_inode = fd_to_inode( fd ) ) ) {
    monitor_read( p_inode, "readv" );
  }
  ret_val = ((int (*)(unsigned long, const struct iovec *, long))orig_readv)
    ( fd, vector, count );

  return( ret_val );
} /* lomac_readv() */



int lomac_fchown( unsigned int fd, uid_t user, gid_t group ) {

  gid_t original_gid;         /* holds gid `fd' had before call */
  struct file *p_file;        /* pointer to file struct for `fd' */
  struct inode *p_inode;      /* pointer to inode struct for `fd' */

  /* Make `p_file' point to the file corresponding to `fd'.  Like *
   * sys_fchown(), we're not going to use fget().  We just want a  *
   * brief atomic peek at the inode, so we're not going to fool   *
   * with all that reference-counting fget() and fput() jazz.     */
  if( fd >= NR_OPEN || !( p_file = current->files->fd[ fd ] ) ) {
    /* sys_fchown() will bomb with -EBADF in this case if *
     * we continued on - we might as well bomb right now. */
    return( -EBADF );
  }

  /* Make `p_inode' point to the inode corresponding to `fd'. */
  if( !( p_inode = p_file->f_dentry->d_inode ) ) {
    /* again, we'll bomb since sys_fchown() will if we don't */
    return( -ENOENT );
  }

  original_gid = p_inode->i_gid;

  /* Will the proposed gid change cause a level change according to    *
   * the GLM?  Eventually, we'll want the FSOLM to add an exception to *
   * the FELM and IELM and allow the chgrp, but for now we'll prevent  *
   * the chgrp from happening.                                         */
  if( will_confuse_glm( original_gid, group ) ) {
    return( -EPERM );
  }

  /* Call kernel's original sys_fchown() function. */
  return( ((int (*)(unsigned int, uid_t, gid_t))orig_fchown)(fd,user,group) );

} /* lomac_fchown() */


int lomac_chown( const char *filename, uid_t user, gid_t group ) {

  gid_t original_gid;         /* holds gid `filename' had before call */
  struct dentry *p_dentry;    /* pointer to dir cache entry for filename */

  /* Put current gid of `filename's file into `original_gid' .*
   * We use lnamei() as the kernel's sys_chown() uses it.     */

  p_dentry = lnamei( filename );  /* get dir cache entry for `filename' */
  if( IS_ERR( p_dentry ) ) {  
    /* if lnamei() fails here, it will also fail when it is called at  *
     * the begining of the kernel's original sys_chown() routine.  So, *
     * we can return without calling the original sys_chown().         */
    return( PTR_ERR( p_dentry ) );
  }
  dput( p_dentry );
  original_gid = p_dentry->d_inode->i_gid;

  /* Will the proposed gid change cause a level change according to    *
   * the GLM?  Eventually, we'll want the FSOLM to add an exception to *
   * the FELM and IELM and allow the chgrp, but for now we'll prevent  *
   * the chgrp from happening.                                         */
  if( will_confuse_glm( original_gid, group ) ) {
    return( -EPERM );
  }

  /* Call kernel's original sys_chown() function. */
  return( ((int (*)(const char *, uid_t, gid_t))orig_chown)( filename,
							     user, group ) );

} /* lomac_chown() */


int lomac_open( const char *filename, int flags, int mode ) {

  int file_already_exists_flag  = 0;
  char *kernel_filename_s;  /* kernel-space copy of `filename' */
  lvnode_t *p_inode;        /* inode for `filename' */
  int ret_val;              /* holds return value */
 


  /* We intercept the open() system call not only to perform mediation, but *
   * to update the FSOLM by assigning levels to newly-created files.   *
   * Since we are interested only in files that are created, and not   *
   * those that are simply reopened, we do some crufty                 *
   * does-this-file-exist checks below.                                */

  kernel_filename_s = getname( filename );
  if( IS_ERR( kernel_filename_s ) ) {
    return( PTR_ERR( kernel_filename_s ) );
  }

  if( ( p_inode = lvnode_from_path( kernel_filename_s ) ) ) {

    file_already_exists_flag = 1;

    /* If this open wants to truncate an existing file, *
     * we've got to do write mediation.                 */
    if( flags & O_TRUNC ) {

      if( monitor_write( p_inode, "open/trunc" ) ) {
	putname( kernel_filename_s );
	return( -EPERM );
      }      

    } /* if truncating */

  } /* if existing file */

  putname( kernel_filename_s );

  ret_val = ((int (*)(const char *, int, int))orig_open)
    ( filename, flags, mode );

  /* If this was a successful open, make sure the ILM is up-to-date. */
  if( ret_val >= 0 ) {

    /* Call the appropriate FSOLM update function depending *
     * on whether or not the file was just created.         */
    if( file_already_exists_flag ) { 
      fsolm_set_for_existing_object( fd_to_inode( ret_val ), 1 );
    } else {
      fsolm_set_for_new_object( fd_to_inode( ret_val ), 1 );
    }

  }

  return( ret_val );

} /* lomac_open() */


int lomac_pipe( unsigned long *fildes ) {

  int ret_val;              /* holds return value */
  unsigned long fd_index;   /* file descriptor index for pipe inode */


  ret_val = ((int (*)(unsigned long *))orig_pipe)( fildes );

  /* If this was a successful pipe creation, inform the VLM is up-to-date. */
  if( !ret_val ) {

    /* The kernel's sys_pipe() call writes the appropriate file      *
     * descriptor values directly out to the calling process's file  *
     * table.  We must copy one of the file descriptors back into    *
     * kernel space to figure out what inode the kernel is using for *
     * the pipe.                                                     */
    copy_from_user( &fd_index, fildes, sizeof( long ) );


    /* This leap of dereferencing faith should work, since the *
     * original syscall just completed successfully.           */
    fsolm_set_for_new_object( fd_to_inode( fd_index ), 0 );

  }

  return( ret_val );

} /* lomac_pipe() */


int lomac_creat( const char *filename, int mode ) {

  int file_already_exists_flag  = 0;
  char *kernel_filename_s;  /* kernel-space copy of `filename' */
  lvnode_t *p_inode;        /* inode for `filename' */
  int ret_val;              /* holds return value */

  /* The sys_creat() system call truncates existing files, *
   * so we must do write mediation.                        */
  kernel_filename_s = getname( filename );
  if( IS_ERR( kernel_filename_s ) ) {
    return( PTR_ERR( kernel_filename_s ) );
  }

  if( ( p_inode = lvnode_from_path( kernel_filename_s ) ) ) {

    file_already_exists_flag = 1;

    if( monitor_write( p_inode, "creat" ) ) {
      putname( kernel_filename_s );
      return( -EPERM );
    }

  } /* if existing file */

  putname( kernel_filename_s );


  ret_val = ((int (*)(const char *, int))orig_creat)( filename, mode );

  /* If this was a successful creat, make sure the ILM is up-to-date. */
  if( ret_val >= 0 ) {

    /* Call the appropriate FSOLM update function depending *
     * on whether or not the file was just created.         */
    if( file_already_exists_flag ) { 
      fsolm_set_for_existing_object( fd_to_inode( ret_val ), 1 );
    } else {
      fsolm_set_for_new_object( fd_to_inode( ret_val ), 1 );
    }

  }

  return( ret_val );

} /* lomac_creat() */



/* FYI - unix protocol code is in net/unix/af_unix.c */

int lomac_socketcall( int call, unsigned long *args ) {

  unsigned long arg_array[ 6 ];   /* holds all args to sys_socketcall() */
  unsigned long arg0;             /* first argument to sys_socketcall() */
  struct inode *p_socket_inode;   /* points to the bound socket's inode */
  int ret_val;                    /* holds return value */
  struct socket *p_socket;        /* points to `p_socket_inode's socket */
  struct sock   *p_sock;          /* points to `p_socket's sock struct  */
  struct inode  *p_other_inode;   /* sockets often work in pairs        */

  /* Make the call to the original sys_socketcall() */
  ret_val = ((int (*)(int, unsigned long *))orig_socketcall)( call, args );

  switch( call ) {

  case SYS_BIND:

    /* If this was a successful bind() call to a UNIX domain socket, *
     * we need to make sure the FSOLM is up-to-date.  Was this a     *
     * successful call to bind()?                                    */
    copy_from_user( arg_array, args, ( sizeof( unsigned long ) * 3 ) );
    arg0 = arg_array[ 0 ];
    p_socket_inode = (current->files->fd[ arg0 ])->f_dentry->d_inode;
    if( ( p_socket_inode->i_sock ) &&
	( p_socket_inode->u.socket_i.ops->family == PF_UNIX ) &&
	( !ret_val ) ) {

      /* The binding has created a name in some filesystem namespace, *
       * represented by its own inode.  Make `p_other_inode' point to *
       * this inode and update the FSOLM in response to its creation. */
      p_socket = &(p_socket_inode->u.socket_i);
      p_sock   = (struct sock *)p_socket->sk;
      p_other_inode = p_sock->protinfo.af_unix.dentry->d_inode;
      handle_unix_domain_bind( p_socket_inode, p_other_inode );

     /* printk( "server bind socket %u:%lu level %u to name %u:%lu level %u\n",
      *	      p_socket_inode->i_dev, p_socket_inode->i_ino,
      *       fsolm_get( p_socket_inode ),
      *	      p_other_inode->i_dev, p_other_inode->i_ino,
      *	      fsolm_get( p_other_inode ) );
      */

    } /* if successful bind() */
    break;

  case SYS_CONNECT:

    copy_from_user( arg_array, args, ( sizeof( unsigned long ) * 3 ) );
    arg0 = arg_array[ 0 ];
    p_socket_inode = (current->files->fd[ arg0 ])->f_dentry->d_inode;
    if( ( p_socket_inode->i_sock ) &&
	( p_socket_inode->u.socket_i.ops->family == PF_UNIX ) &&
	( !ret_val ) ) {
    
      p_socket = &(p_socket_inode->u.socket_i);
      p_sock   = (struct sock *)p_socket->sk;

      /* sys_connect() links the client's socket to the socket      *
       * created by sys_accept() in the server.  The implementation *
       * of the link has changed a bit from the 2.0 Linux kernels.  */
      p_other_inode = p_sock->pair->protinfo.af_unix.dentry->d_inode;

      /* printk( "pid %u socket %lu, %u connected to socket %lu, %u\n",
       *      current->pid, p_socket_inode->i_ino, p_socket_inode->i_dev,
       *      p_other_inode->i_ino, p_other_inode->i_dev );
       */

      handle_unix_domain_accept_connect( p_other_inode, p_socket_inode );

    /* printk( "client connect sock %u:%lu level %u to serv %u:%lu level %u\n",
     *	      p_socket_inode->i_dev, p_socket_inode->i_ino,
     *	      fsolm_get( p_socket_inode ),
     *	      p_other_inode->i_dev, p_other_inode->i_ino,
     *	      fsolm_get( p_other_inode ) );
     */

    }

    break;

  case SYS_ACCEPT:

    /* If this was a successful accept() call to a UNIX-domain socket, *
     * then the kernel has created a new socket for handling the       *
     * accepted connection.  We must make `p_other_inode' point to     *
     * this new socket's inode, and update the ILM in response to its  *
     * creation.                                                       */
    copy_from_user( arg_array, args, ( sizeof( unsigned long ) * 3 ) );
    arg0 = arg_array[ 0 ];
    p_socket_inode = (current->files->fd[ arg0 ])->f_dentry->d_inode;
    if( ( p_socket_inode->i_sock ) &&
	( p_socket_inode->u.socket_i.ops->family == PF_UNIX ) &&
	( ret_val >= 0 ) ) {

      p_other_inode = (current->files->fd[ ret_val ])->f_dentry->d_inode;

      /* printk( "pid %u will handle connection to %lu, %u with %lu, %u\n",
       *      current->pid,
       *      p_socket_inode->i_ino, p_socket_inode->i_dev,
       *      p_other_inode->i_ino, p_other_inode->i_dev );
       */

      handle_unix_domain_accept_connect( p_socket_inode, p_other_inode );

     /* printk( "server accept sock %u:%lu level %u to serv %u:%lu level %u\n",
      *	      p_socket_inode->i_dev, p_socket_inode->i_ino,
      *	      fsolm_get( p_socket_inode ),
      *	      p_other_inode->i_dev, p_other_inode->i_ino,
      *	      fsolm_get( p_other_inode ) );
      */

    }
    break;

  } /* switch( call ) */

  return( ret_val );

} /* lomac_socketcall() */


int lomac_kill( int pid, int sig ) {

  int ret_val;   /* holds function return values */

  if( ( ret_val = monitor_kill( pid, sig ) ) ) {
    return( ret_val );
  }
  return( ((int (*)(int, int))orig_kill)( pid, sig ) );

} /* lomac_kill() */


int lomac_write( unsigned int fd, char *buf, unsigned int count ) {

  struct inode *p_inode;              /* inode corresponding to `fd' */

  if( ( p_inode = fd_to_inode( fd ) ) ) {
    if( monitor_write( p_inode, "write" ) ) {
      return( -EPERM );
    }
  }

  return( ((int (*)(unsigned int, char *, int))orig_write)( fd, buf, count ) );

} /* lomac_write() */


int lomac_writev( unsigned long fd, const struct iovec *vector, long count ) {

  struct inode *p_inode;              /* inode corresponding to `fd' */

  if( ( p_inode = fd_to_inode( fd ) ) ) {
    if( monitor_write( p_inode, "writev" ) ) {
      return( -EPERM );
    }
  }

  return( ((int (*)(unsigned long, const struct iovec *, long))orig_writev)
    ( fd, vector, count ) );

} /* lomac_writev() */


int lomac_unlink( const char *pathname ) {

  struct dentry *p_dentry;  /* dirent cache entry for `pathname' */
  char *tmp;  /* temporary kernel-space storage for `pathname' */

  tmp = getname( pathname );
  if( ! IS_ERR( tmp ) ) {
    p_dentry = open_namei( tmp, 0x00, 0x00 );
    putname( tmp );
    if( ! IS_ERR( p_dentry ) ) {
      dput( p_dentry );
      if( monitor_write( p_dentry->d_inode, "unlink" ) ) {
	return( -EPERM );
      }
    }
  }

  return( ((int (*)(const char *))orig_unlink)( pathname ) );

} /* lomac_unlink() */


int lomac_rmdir( const char *pathname ) {

  struct dentry *p_dentry;  /* dirent cache entry for `pathname' */
  char *tmp;    /* temporary kernel-space storage for `pathname' */


  /* We must ensure that the current process can write the directory  *
   * specified by `pathname' before we allow it to rmdir it.  If we   *
   * encounter an error while making our check for write permission,  *
   * we will not report it ourselves.  Instead, we will call the      *
   * kernel's original sys_rmdir() function and let it encounter and  *
   * handle the error itself.                                         */

  tmp = getname( pathname );
  if( !IS_ERR( tmp ) ) {
    p_dentry = open_namei( tmp, 0x00, 0x00 );
    putname( pathname );
    if( !IS_ERR( p_dentry ) ) {
      dput( p_dentry );
      if( monitor_write( p_dentry->d_inode, "rmdir" ) ) {
	putname( tmp );
	return( -EPERM );
      }
    }
  }

  putname( tmp );
  return( ((int (*)(const char *))orig_rmdir)( pathname ) );

} /* lomac_rmdir() */


int lomac_truncate( const char *path, unsigned long length ) {

  struct dentry *p_dentry;  /* dirent cache entry for `path' */
  char *tmp;    /* temporary kernel-space storage for `path' */

  /* We must ensure that the current process can write the file      *
   * specified by `path' before we allow it to truncate it.  If we   *
   * encounter an error while making our check for write permission, *
   * we will not report it ourselves.  Instead, we will call the     *
   * kernel's original sys_truncate() function and let it encounter  *
   * and handle the error itself.                                    */

  tmp = getname( path );
  if( !IS_ERR( tmp ) ) {
    p_dentry = open_namei( tmp, 0x00, 0x00 );
    putname( path );
    if( !IS_ERR( p_dentry ) ) {
      dput( p_dentry );
      if( monitor_write( p_dentry->d_inode, "truncate" ) ) {
	putname( tmp );
	return( -EPERM );
      }
    }
  }

  putname( tmp );
  return( ((int (*)(const char *, unsigned long))orig_truncate)
	  ( path, length ) );

} /* lomac_truncate() */


int lomac_ftruncate( unsigned int fd, unsigned long length ) {

  struct inode *p_inode;  /* inode corresponding to `path' */

  if( ( p_inode = fd_to_inode( fd ) ) ) {
    if( monitor_write( p_inode, "ftruncate" ) ) {
      return( -EPERM );
    }
  }

  return( ((int (*)(unsigned int, unsigned long))orig_ftruncate)
	  ( fd, length ) );

} /* lomac_ftruncate() */


int lomac_setpgid( pid_t pid, pid_t pgid ) {

  int ret_val;                       /* holds function return values */
  lprocess_t *p_changing_process;  /* process whose pgid is changing */
  lpid_t old_pgrp;     /* old pgrp of process whose pgrp is changing */

  /* setpgid() is used for two purposes:
   * 1.  Set a process into its own, newly created, pgrp.
   * 2.  Set a process into an existing pgrp.
   *
   * In the first case, we want to add the new pgrp into the PLM.
   * In the second case, we want to ensure that the change in pgrp
   * will not result in the changing process gaining a higher level.
   */

  /* Make `p_changing_process' point to the process whose pgrp may be   *
   * changed.  We identify this process based on the contents of `pid': *
   * IF pid == 0
   *  we want the `current' process,
   * ELSE
   *  we want the process with pid `pid'.
   */

  if( pid ) {
    if( !( p_changing_process = lprocess_from_pid( pid ) ) ) {
      return( -ESRCH );
    }
  } else {
    p_changing_process = lprocess_current();
  }

  /* Save `p_changing_process's current pgid in `old_pgrp'. */
  old_pgrp = lprocess_pgrp( p_changing_process );

  /* If we're changing `p_changing_process' to another existing pgrp,  *
   * we must make sure the change won't result in a level increase for *
   * `p_changing_process'.                                             */
  if( pgid ) {

    if( monitor_pgrp_change( p_changing_process, old_pgrp, pgid ) ) {
      return( -EPERM );
    }

    /* Perform pgrp-transfer revocation of objects on p_changing_process *
     * prob'ly should be done in monitor_pgrp_change, pass extra parm.   */

  } /* if changing to an existing pgrp */


  /* Either we're creating a new pgrp, or making an allowable change *
   * between existing pgrps.  Call to the kernel's setpgid().        */
  if( !( ret_val = ((int (*)(pid_t, pid_t))orig_setpgid)( pid, pgid ) ) ) {

    /* If the call to setpgid() created a new pgrp, add it to the PLM. */

    if( ( !pgid ) || ( pgid == lprocess_pid( p_changing_process ) ) ) {
      update_plm_for_new_pgrp( old_pgrp, lprocess_pid( p_changing_process ) );
    }

  } /* If the kernel's setpgid() succeeded. */

  return( ret_val );
  
} /* lomac_setpgid() */


int lomac_setsid( void ) {

  int ret_val;
  lpid_t old_pgrp;     /* old pgrp of process whose pgrp is changing */

  old_pgrp = lprocess_pgrp( lprocess_current() );

  if( ( ret_val = ((int (*)( void ))orig_setsid)() ) ) {

    update_plm_for_new_pgrp( old_pgrp, current->pid );

  }

  return( ret_val );

} /* lomac_setsid */
    

/* ------ The following system calls require privilege checks only ------ */

int lomac_reboot( int magic, int magic_too, int flag ) {

  if( !has_sufficient_privilege() ) {
    return( -EPERM );
  }

  return( ((int (*)(int, int, int))orig_reboot)( magic, magic_too, flag ) );

} /* lomac_reboot() */


int lomac_swapon( const char *specialfile, int swap_flags ) {

  if( !has_sufficient_privilege() ) {
    return( -EPERM );
  }

  return( ((int (*)(const char *, int))orig_reboot)
	  ( specialfile, swap_flags ) );

} /* lomac_swapon() */


int lomac_swapoff( const char *specialfile ) {

  if( !has_sufficient_privilege() ) {
    return( -EPERM );
  }

  return( ((int (*)(const char *))orig_reboot)( specialfile ) );

} /* lomac_swapoff() */


