/************************************************************************
 *
 * lomac_elm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * This file contains the implementation of the Exception
 * Level Map (ELM).  The ELM is a component of the FSOLM.  See
 * lomac_fsolm.c for more info.
 *
 ************************************************************************/


#include "lomac_elm.h"

#ifdef PARANOID
#include "lomac_rta.h"
#endif


#ifdef PARANOID
#define ELM_LISTNODE_MAGIC    0xBeefFeed
#endif


typedef struct elm_listnode {
#ifdef PARANOID
  unsigned int magic;             /* magic number for paranoid assertions */
#endif
  char   *filename_s;             /* pointer to filename string */
  level_t level;                  /* level associated with `filename_s' */
  struct elm_listnode *p_next;    /* pointer to next node in list */
} elm_listnode_t;

elm_listnode_t *p_elm_list_head;  /* head of elm_listnode list */


/* ------------------------- Private Functions -------------------------- */

#ifdef PARANOID

int is_malformed_elm_listnode( elm_listnode_t * );

/* is_malformed_elm_listnode()
 *
 * in:     p_node - potentially bad pointer to potentially malformed elm
 *                  listnode
 * out:    nothing
 * return: 1 if `p_node' is not a well-formed elm listnode, otherwise 0.
 *
 * This function is a convenient way to make sure a given elm_listnode_t
 * pointer points to a valid node.
 * */

int is_malformed_elm_listnode( elm_listnode_t *p_node ) {

  if( !p_node ) {
    return( 1 );
  }

  if( p_node->magic != ELM_LISTNODE_MAGIC ) {
    return( 1 );
  }

  if( is_malformed_absolute_path( p_node->filename_s ) ) {
    return( 1 );
  }

  if( ( p_node->level < LOMAC_LOWEST_LEVEL ) ||
      ( p_node->level > LOMAC_HIGHEST_LEVEL ) ) {
    return( 1 );
  }

  return( 0 );
} /* is_malformed_elm_listnode() */


#endif


/* ------------------------- Public Functions -------------------------- */


/* init_elm()
 *
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 */

void init_elm( void ) {

  p_elm_list_head = NULL;
  
  elm_set( "/dev/log", LOMAC_LOWEST_LEVEL );
  elm_set( "/dev/low", LOMAC_LOWEST_LEVEL );
  elm_set( "/var/log/httpd/access_log", LOMAC_LOWEST_LEVEL );

#ifdef SERVER_POLICY
  elm_set( "/dev/ttyS1", 2 );
#else
  /* elm_set( "/dev/null", LOMAC_LOWEST_LEVEL ); */
  elm_set( "/dev/psaux", LOMAC_LOWEST_LEVEL );
  elm_set( "/tmp/.X11-unix/X0", LOMAC_LOWEST_LEVEL );
  elm_set( "/tmp/.X11-unix/X1", LOMAC_HIGHEST_LEVEL );
  elm_set( "/tmp/lowfile", LOMAC_LOWEST_LEVEL );
  elm_set( "/tmp/highfile", LOMAC_HIGHEST_LEVEL );
#endif

  /* elm_set( "/etc/wtmplock", LOMAC_LOWEST_LEVEL ); */
  /* elm_set( "/var/log/wtmp", LOMAC_LOWEST_LEVEL );  */
  /* elm_set( "/var/run/utmp", LOMAC_LOWEST_LEVEL );  */

} /* init_elm() */


/* elm_set()
 *
 * in:     filename_s
 *         level
 * out:    nothing
 * return: nothing
 *
 *
 */

void elm_set( char *filename_s, level_t level ) {

  elm_listnode_t *p_node;

#ifdef PARANOID
  if( is_malformed_absolute_path( filename_s ) ) {
    panic( "LOMAC(lomac_elm.c) malformed path string" );
  }

  if( ( level < LOMAC_LOWEST_LEVEL ) || ( level > LOMAC_HIGHEST_LEVEL ) ) {
    panic( "LOMAC(lomac_elm.c) level out of bounds" );
  }
#endif

  /* Search for an existing ELM mapping for `filename_s'. *
   * If found, overwrite the old level with the new one.  */
  for( p_node = p_elm_list_head; p_node; p_node = p_node->p_next ) {

#ifdef PARANOID
    if( is_malformed_elm_listnode( p_node ) ) {
      panic( "LOMAC(lomac_elm.c) malformed node in ELM" );
    }
#endif

    if( ! strcmp( p_node->filename_s, filename_s ) ) {
      p_node->level = level;

#ifdef PARANOID
      if( is_malformed_elm_listnode( p_node ) ) {
	panic( "LOMAC(lomac_elm.c) created malformed node in ELM" );
      }
#endif

      return;
    }

  } /* for all mappings in the ELM list */

  /* If we arrive here, no mapping for `filename_s' exists *
   * in the ELM.  We must add one.                         */

  p_node = (elm_listnode_t *)lmalloc( sizeof( elm_listnode_t ) );

#ifdef PARANOID
  if( !p_node ) {
    panic( "Lomac(lomac_elm.c) - unable to allocate new ELM listnode" );
  }
#endif

  p_node->filename_s = filename_s;
  p_node->level      = level;
  p_node->p_next     = p_elm_list_head;
  p_elm_list_head    = p_node;

#ifdef PARANOID
  p_node->magic = ELM_LISTNODE_MAGIC;

  if( is_malformed_elm_listnode( p_node ) ) {
    panic( "lomac(lomac_elm.c) - created malformed ELM node" );
  }
#endif

} /* elm_set() */


/* elm_get()
 *
 * in:     p_lvnode_id - unique id of inode for which we want to search
 *                       for a level exception
 * out:    nothing
 * return: the level indicated by an exception entry if one is found, else 0.
 *
 *     This function searches the ELM for an exception entry corresponding
 * to `p_lvnode_id'.  If it finds one, it returns the level indicated by
 * the exception entry.  If it does not find one, it returns 0.  Note that
 * the ELM stores entries by filename, not by unique lvnode id, since at
 * any given time, some of the files corresponding to ELM entries may not
 * exist, and therefore have no lvnode.  The fact that the ELM stores
 * entries by name means that every query (that is, call to elm_get())
 * is handled by searching through the ELM entry list, translating each
 * name into an lvnode for comparison.
 *
 */

level_t elm_get( lvnode_id_t *p_lvnode_id ) {

  elm_listnode_t *p_elm_entry;   /* used to traverse ELM list */
  lvnode_t *p_lvnode;            /* lvnodes corresponding to ELM entries */
  lvnode_id_t lvnode_id;         /* unique id of `p_lvnode' */


  for( p_elm_entry = p_elm_list_head; p_elm_entry; 
       p_elm_entry = p_elm_entry->p_next ) {

    if( ( p_lvnode = lvnode_from_path( p_elm_entry->filename_s ) ) ) {

      lvnode_to_id( p_lvnode, &lvnode_id );
      if( lvnode_equals( &lvnode_id, p_lvnode_id ) ) {

	return( p_elm_entry->level );

      } /* if ELM entry matches `p_lvnode_id' */

    } /* if ELM entry corresponds to an existing file */

  } /* for all ELM entries */

  return( 0 );  /* no entry found in ELM */

} /* elm_get() */





