/*************************************************************************
 *
 * lomac_glm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 *
 * GLM is assumes that any gid it doesn't know about is mapped to
 * LOMAC_LOWEST_LEVEL.  I assume that most, if not all, of the high-level
 * groups like "root" will be created before LOMAC is initialized.
 * Furthermore, most groups created during run-time will be low-level
 * user-oriented groups.  Therefore, the LOMAC_LOWEST_LEVEL default
 * makes more sense than a LOMAC_HIGHEST_LEVEL default.
 *
 *************************************************************************/

#include "kernel_kernel.h"
#include "kernel_lgid.h"

#include "lomac_glm.h"

#ifdef PARANOID
#define GLM_HASHNODE_MAGIC   0xAceCabAd
#endif
#define GLM_HASHNUM          0x1f
#define GLM_HASH( gid )      ( (gid) & GLM_HASHNUM )

typedef struct glm_hashnode {
#ifdef PARANOID
  unsigned int magic;           /* magic number for paranoid assertions */
#endif
  lgid_t   gid;                 /* group id */
  level_t level;                /* level equivalent to group id `gid' */
  struct glm_hashnode *p_next;  /* next bucket in hash chain */
} glm_hashnode_t;

glm_hashnode_t *glm_hash_table[ GLM_HASHNUM + 1 ];

/* gid_to_level()
 *
 * in:     gid - gid whose level we wish to know
 * out:    nothing
 * return: level of `gid'
 * 
 * Search the GLM for a mapping from `gid' to a level.  If we find a 
 * mapping, return the indicated level.  If we find no mapping, we
 * assume the correct level is LOMAC_LOWEST_LEVEL and return that.
 *
 */

level_t gid_to_level( lgid_t gid ) {

  glm_hashnode_t *p_hashnode;  /* points to hashnodes in the GLM hash table */

  p_hashnode = glm_hash_table[ GLM_HASH( gid ) ];
  while( p_hashnode ) {

#ifdef PARANOID
    if( p_hashnode->magic != GLM_HASHNODE_MAGIC ) {
      panic( "Lomac (lomac_glm.c) : Bad magic number on GLM hashnode query" );
    }
    if( ( p_hashnode->level < LOMAC_LOWEST_LEVEL ) ||
	( p_hashnode->level > LOMAC_HIGHEST_LEVEL ) ) {
      panic( "LOMAC(lomac_glm.c) Out of bounds level in GLM" );
    }
#endif
      
    if( p_hashnode->gid == gid ) {
      return( p_hashnode->level );
    }

    p_hashnode = p_hashnode->p_next;

  } /* for all hashnodes in chain */

  /* If we reach here, then `gid' does not exist in the GLM.  In  *
   * this case we assume the proper level is LOMAC_LOWEST_LEVEL. */
  return( LOMAC_LOWEST_LEVEL );

} /* gid_to_level() */


/* set_gid_level()
 *
 * in:     gid   - gid whose level we want to set
 *         level - level to set `gid' to
 * out:    nothing
 * return: nothing
 *
 *     Sets the level of `gid' to `level' in the GLM.
 *
 */

void set_gid_level( lgid_t gid, level_t level ) {

  glm_hashnode_t *p_hashnode;  /* points to hashnodes in the GLM hash table */
  int hash_chain_index;        /* holds results of GLM_HASH() */

#ifdef PARANOID
  if( ( level < LOMAC_LOWEST_LEVEL ) || ( level > LOMAC_HIGHEST_LEVEL ) ) {
    panic( "LOMAC(lomac_glm.c) level out of bounds" );
  }
#endif

  /* First, try to find an existing entry for `gid' in *
   * the GLM hash table and update its level field.    */
  
  hash_chain_index = GLM_HASH( gid );
  p_hashnode = glm_hash_table[ hash_chain_index ];
  while( p_hashnode ) {

#ifdef PARANOID
    if( p_hashnode->magic != GLM_HASHNODE_MAGIC ) {
      panic( "Lomac (lomac_glm.c) : Bad magic number on GLM hashnode set" );
    }
#endif
      
    if( p_hashnode->gid == gid ) {
      p_hashnode->level = level;   /* a node already exists - update it. */

#ifdef PARANOID
      if( level != gid_to_level( gid ) ) {
	panic( "LOMAC(lomac_glm.c) failed to set gid level" );
      }
#endif

      return;
    }

    p_hashnode = p_hashnode->p_next;

  } /* for all hashnodes in chain */

  /* If we reach here, then an entry for `gid' does not exist in *
   * the GLM.  we must add a new hashnode to the GLM for `gid'   *
   * at the head of the chain indicated by `hash_chain_index'.   */

  p_hashnode = (glm_hashnode_t *)lmalloc( sizeof( glm_hashnode_t ) );
#ifdef PARANOID
  if( !p_hashnode ) {
    panic( "Lomac (lomac_glm.c) : Unable to allocate new GLM hashnode" );
  }
  p_hashnode->magic = GLM_HASHNODE_MAGIC;
#endif

  p_hashnode->gid = gid;
  p_hashnode->level = level;
  p_hashnode->p_next = glm_hash_table[ hash_chain_index ];
  glm_hash_table[ hash_chain_index ] = p_hashnode;

#ifdef PARANOID
  if( level != gid_to_level( gid ) ) {
    panic( "LOMAC(lomac_glm.c) failed to set gid level" );
  }
#endif


} /* set_gid_level() */


/* init_glm()
 *
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 *
 */

void init_glm( void ) {

  /* before we begin, clear `glm_hash_table' */
  memset( (void *)glm_hash_table, 0,
	  ( sizeof( glm_hashnode_t *) * GLM_HASHNUM + 1 ) );

  /* set levels for some gids */
  set_gid_level(  0, LOMAC_HIGHEST_LEVEL );  /* root */
  set_gid_level(  1, LOMAC_HIGHEST_LEVEL );  /* bin */
  set_gid_level(  2, LOMAC_HIGHEST_LEVEL );  /* daemon */
  set_gid_level(  3, LOMAC_HIGHEST_LEVEL );  /* sys */
  set_gid_level(  4, LOMAC_HIGHEST_LEVEL );  /* adm */
  set_gid_level(  5, LOMAC_HIGHEST_LEVEL );  /* tty */
  set_gid_level(  6, LOMAC_HIGHEST_LEVEL );  /* disk */
  set_gid_level(  7, LOMAC_HIGHEST_LEVEL );  /* lp */
  set_gid_level(  8, LOMAC_HIGHEST_LEVEL );  /* mem */
  set_gid_level(  9, LOMAC_HIGHEST_LEVEL );  /* kmem */
  set_gid_level( 10, LOMAC_HIGHEST_LEVEL );  /* wheel */
  set_gid_level( 12, LOMAC_HIGHEST_LEVEL );  /* mail */
  set_gid_level( 13, LOMAC_HIGHEST_LEVEL );  /* news */
  set_gid_level( 14, LOMAC_HIGHEST_LEVEL );  /* uucp */
  set_gid_level( 15, LOMAC_HIGHEST_LEVEL );  /* man */
  set_gid_level( 16, LOMAC_HIGHEST_LEVEL );  /* utmp on some systems */
  set_gid_level( 19, LOMAC_HIGHEST_LEVEL );  /* floppy */
  set_gid_level( 20, LOMAC_HIGHEST_LEVEL );  /* games */
  set_gid_level( 21, LOMAC_HIGHEST_LEVEL );  /* slocate */
  set_gid_level( 30, LOMAC_HIGHEST_LEVEL );  /* gopher */
  set_gid_level( 40, LOMAC_HIGHEST_LEVEL );  /* dip */
  set_gid_level( 50, LOMAC_HIGHEST_LEVEL );  /* ftp */
  set_gid_level( 101, LOMAC_HIGHEST_LEVEL );  /* console */
  set_gid_level( 102, LOMAC_HIGHEST_LEVEL );  /* utmp on some systems */
  set_gid_level( 233, LOMAC_HIGHEST_LEVEL );  /* xfs */


  set_gid_level(  99, LOMAC_LOWEST_LEVEL );  /* nobody */
  set_gid_level( 100, LOMAC_LOWEST_LEVEL );  /* users */
  set_gid_level( 230, LOMAC_LOWEST_LEVEL );  /* pppusers */
  set_gid_level( 231, LOMAC_LOWEST_LEVEL );  /* popusers */
  set_gid_level( 232, LOMAC_LOWEST_LEVEL );  /* slipusers */


} /* init_glm() */
