/****************************************************************************
 *
 * lomac_monitor.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * The lomac reference monitor.  Each function in this file exists for
 * one or both of the following two purposes:
 *
 * 1.  To update LOMAC data structures such as the PLM and FSOLM
 *     whenever the kernel does something important, in order to
 *     ensure that these data structures always accurately relfect
 *     the current state of the kernel.
 *
 * 2.  To make an access control decision whenever a user-space
 *     process asks the kernel to perform some security-relevant
 *     operation (like reading or writing a file) and to prevent the
 *     kernel from performing the operation if it would violate
 *     LOMAC's security policy.  The combination of decision making
 *     and possible prevention is called "mediation".
 *
 * All of these functions are called from various replacement system
 * call functions found in lomac_syscalls.c .  Thanks to the system
 * call table patching accomplished by the code in linux_lkm.c, these
 * replacement system call functions are called before the kernel
 * executes certain important system calls.  The replacement system
 * calls act as an abstraction barrier, hiding kernel-specific
 * details.  They call the functions in this file to do the
 * LOMAC-specific work that is platform independent.  Thus, these
 * functions can be thought of as "handlers" or "callbacks" for system
 * call events.
 *
 * For a given system call, LOMAC performs mediation by comparing the
 * level of the subject making the system call to the level of the
 * object the system call operates upon.  There is only one kind of
 * subject in the kerel: the process group.  The LOMAC reference
 * monitor determines the level of a given process group using the
 * PLM.  There are numerous kinds of objects in the kernel.  The LOMAC
 * reference monitor determines their levels as follows:
 *
 * object                                  map    level can change?  
 * --------------------------------------- -----  -----------------------
 * filesystem entities (files, dirs, etc.) FSOLM  no
 * pipes, named (FIFOS)                    FSOLM  no
 * pipes, unnamed                          FSOLM  may drop with writes
 * message queues                          FSOLM  no
 * sockets, PF_UNIX                        FSOLM  no
 * sockets, other                          NLM    changes with source NIC
 *
 * In addition to these objects, the kernel provides additional
 * abstractions which LOMAC does not consider objects: semaphores and
 * shared memory segments.  The LOMAC reference monitor does not
 * mediate accesses to these abstractions.  Instead, it ensures that
 * these objects may be shared only between processes in the same
 * process group - that is, they may not be shared between subjects,
 * and are therefore harmless.
 *
 * In addition to performing mediation on object accesses, the LOMAC
 * reference monitor also performs privilege checks on certain
 * critical system calls that do not access objects, such as
 * sys_reboot().  These privilege checks ensure that only subjects at
 * the highest level can access these critical system calls.
 *
 *
 * References:
 *
 * McKusick et. al. `The Design and Implementation of the 4.4 BSD
 * Operating System,' Addison-Wesley 1996, index entries for "FIFO
 * file".
 *
 * National Computer Security Center `Final Evaluation Report TIS
 * Trusted XENIX 3.0' 8 April 1992, Section 2.7.2 TCB Protected
 * Resources - Objects.  (Available by snail-mail from the NCSC.)
 *
 * David A. Rusling `The Linux Kernel' chapter 5: Interprocess
 * Communication Mechanisms, March 1998. (A description of the linux
 * kernel, available on the web.)
 *
 * 
 ****************************************************************************/



#include "kernel_kernel.h"         /* for lpanic()   */
#include "kernel_lgid.h"           /* for lgid_t     */
#include "kernel_lvnode.h"         /* for lvnode_t   */
#include "kernel_lpid.h"           /* for lpid_t     */
#include "kernel_lprocess.h"       /* for lprocess_t */
#include "kernel_lnic.h"

#include "lomac_level.h"
#include "lomac_monitor.h"
#include "lomac_plm.h"
#include "lomac_nlm.h"
#include "lomac_fsolm.h"
#include "lomac_tpl.h"




/* monitor_read()
 *
 * in:     p_inode - inode being read
 *         label_s - string containing name of syscall used for read
 * out:    nothing
 * return: nothing
 *
 * This function determines the level of the object indicated by
 * `p_inode'.  In order to accomplish this feat, it must first
 * determine what kind of object `p_inode' refers to, and then query
 * the level mapping appropriate to that kind of object.
 *
 * Since this function is apt to be called on system calls such as
 * read() before the kernel's sys_read() function is called, if any of
 * the parameters specified by the calling process are erroneous, this
 * function will encounter the errors before sys_read() does.  Rather
 * than duplicate the error-handling behavior of the kernel's
 * system-call functions, this function quietly returns 0 on error.
 * When this function returns 0, the caller should proceed with the
 * call to the kernel's system-call function confident that the kernel
 * will also detect the error and bomb out without actually performing
 * any operation on the object.
 *
 */

void monitor_read( lvnode_t *p_inode, char *label_s ) {

  level_t inode_level = LOMAC_INVALID_LEVEL; /* holds level of `p_inode' */
  level_t proc_level  = LOMAC_INVALID_LEVEL; /* holds level of current proc */
  trust_t proc_trust  = e_invalid;           /* current proc's trust setting */
  lprocess_t *p_current;    /* pointer to the current */
  lpid_t      current_pid;  /* current process's pid */
  char *objname_s = NULL;   /* will hold name of object that caused demotion */

#ifdef VERBOSELOG
  lvnode_id_t lvnode_id;    /* holds unique id of lvnode that cause demotion */
  char *procname_s = NULL;  /* holds name of current exe based on argv[ 0 ] */
#endif

#ifdef PARANOID
  if( !p_inode ) {
    lpanic( "LOMAC(monitor_read) - null inode pointer" );
  }
  if( !label_s ) {
    lpanic( "LOMAC(monitor_read) - null label string" );
  }
#endif


  /*
   *
   * STEP 0:  Determine the level of the current process.  If its
   *          level is LOMAC_LOWEST_LEVEL, then return immediately
   *          since it can't be made any lower no matter what the
   *          process reads.
   *
   */

  p_current = lprocess_current();
  current_pid = lprocess_pgrp( p_current );
  if( plm_get( current_pid, &proc_level, &proc_trust ) ) {
    lpanic( "LOMAC(lomac_monitor.c) - monitor_read got pid without level" );
  }

  /* An optimization - if the process is already at the lowest level, *
   * then there's no point in worrying about demoting it any further. */
  if( proc_level == LOMAC_LOWEST_LEVEL ) {
    return;
  }

  /* If the current process is trusted e_neverdemote, then *
   * there is no point in worrying about demotion, either. */
  if( proc_trust == e_neverdemote ) {
    return;
  }


  /*
   *
   * STEP 1:  Determine the level of `p_inode'.  If `p_inode' represents
   *          a non-local domain socket, we must examine its queue of
   *          incoming network packets.  Otherwise, we can determine its
   *          level according to the FSOLM.
   *
   */

  if( lvnode_is_nonlocal_socket( p_inode ) ) {

    /* `p_inode' represents a socket used for network communication. *
     * As described in lomac_nlm.c, the level associated with a      *
     * network socket may change with each read, depending on the    *
     * level of the NIC which originated the next sockbuf to be      *
     * read.  Return the level associated with this socket for the   *
     * next read (that is, the read being handled by the current     *
     * system call).                                                 */
    if( !( lvnode_network_socket_to_level( p_inode, 
					   &inode_level, &objname_s ) ) ) {
#ifdef VERBOSELOG
      printk( KERN_INFO 
	      "LOMAC(lomac_monitor.c) pid %u read to empty net socket\n",
	      current_pid );
#endif

#ifdef SERVER_POLICY
      return;
#else
      /* For now, we'll assume the worst in this case *
       * and claim the level was the lowest possible. */
      inode_level = LOMAC_LOWEST_LEVEL;
      objname_s = "NIC?";
#endif
    }

  } else {

    /* For now, we'll assign levels to local sockets, unnamed *
     * and named pipes (FIFOs) and files based on the FSOLM.  */
    inode_level = fsolm_get( p_inode );

  } /* if/else `p_inode' is a non-local domain socket */


  /*
   *
   * STEP 2:  Compare the level of `p_inode' to the level of the current
   *          process, and demote the process if necessary.
   *
   */

  if( inode_level < proc_level ) {

    /* demote the process */
    plm_set( current_pid, inode_level, proc_trust );

#ifdef VERBOSELOG

    /* We're going to downgrade `pid's level, so we must construct  *
     * a meaningful log message.  Make `procname_s' point to a      *
     * kernel-space copy of the `current' process's `argv[0]'.      *
     * argv[0] should contain the name of the program the `current' *
     * process is running - we'll use it in the log message.        */

    procname_s = lprocess_getname( p_current );

    if( lvnode_is_nonlocal_socket( p_inode ) ) {
	
      printk( KERN_INFO "process %u (%s) demoted from level %u to %u"
	      " on %s of NIC %s\n",
	      current_pid, procname_s, proc_level, inode_level,
	      label_s, objname_s );
    } else {

      lvnode_to_id( p_inode, &lvnode_id );
      printk( KERN_INFO "process %u (%s) demoted from level %u to %u"
	      " on %s of inode number %lu device %u\n",
	      current_pid, procname_s, proc_level, inode_level,
	      label_s, lvnode_id.lvnode_number, 
	      lvnode_id.device_number );

    }

#endif

    /* free kernel memory allocated for `procname_s' */
    lprocess_putname( procname_s );

  } /* if demotion required */

} /* monitor_read() */

  

/* monitor_kill()
 *
 * in:      pid - pid parm to sys_kill system call.  0 means signal to
 *                the current process's process group, -1 means signal
 *                all processes, and positive numbers means signal a
 *                particular process by pid.
 *          sig - signal number to send
 * out:     nothing
 * return:  0 if sys_kill should proceed, else -EPERM.
 *
 * Monitors the sys_kill system call.
 *
 */

int monitor_kill( int pid, int sig ) {

  lprocess_t *p_current;       /* points to current process */
  lpid_t  current_pgid;         /* pid of current process */
  level_t current_level, other_level;  /* level of current, other process */
  trust_t current_trust, other_trust;  /* turst for current, other process */

#ifdef VERBOSELOG
  char *procname_s = NULL;  /* holds name of current exe based on argv[ 0 ] */
#endif

  p_current = lprocess_current();
  current_pgid = lprocess_pgrp( p_current );

  if( pid == -1 ) {

    /* Signal all processes.  We will allow this only if *
     * the `current' process is at the highest level.    */
    if( plm_get( current_pgid, &current_level, &current_trust ) ) {
      lpanic( "LOMAC(lomac_monitor.c) - monitor_kill(-1) got src PLM miss" );
    }
    if( LOMAC_HIGHEST_LEVEL != current_level ) {
      
#ifdef VERBOSELOG
      procname_s = lprocess_getname( p_current );
      printk( KERN_INFO "process %u (%s) group %u "
	      "denied signal %u to all processes.\n",
	      lprocess_pid( p_current ),procname_s, 
	      current_pgid, sig );
      lprocess_putname( procname_s );
#endif

      return( -EPERM );
    }


  } else if( pid < 0 ) {

    /* Signal all processes in process group number -`pid'. We will  *
     * allow this only if the `current' process has a level as least *
     * as great as the one assigned to the  -`pid' process group.    */

    if( plm_get( current_pgid, &current_level, &current_trust ) ) {
      lpanic( "LOMAC(lomac_monitor.c) - monitor_kill(<) got src PLM miss" );
    }
    if( plm_get( -pid, &other_level, &other_trust ) ) {
      /* lpanic( "LOMAC(lomac_monitor.c) - "  *
       * "monitor_kill(<) got dst PLM miss"); */
      /* assume the user process specified a non-existant pgrp. */
      return( -EPERM );
    } 
    if( current_level < other_level ) {

#ifdef VERBOSELOG
      procname_s = lprocess_getname( p_current );
      printk( KERN_INFO "process %d (%s) group %u "
	      "denied signal %u to pgrp %d.\n",
	      lprocess_pid( p_current ), procname_s, current_pgid,
	      sig, -pid );
      lprocess_putname( procname_s );
#endif

      return( -EPERM );

    } 

  } else { /* pid > 0 */

    /* Signal process with pid `pid'.  We will allow this only if the *
     * `current' process's level is at least as great as `pid's.      */
    if( plm_get( current_pgid, &current_level, &current_trust ) ) {
      lpanic( "LOMAC(lomac_monitor.c) - monitor_kill got src PLM miss" );
    }
    if( plm_get( lprocess_pgrp( lprocess_from_pid( pid ) ), 
		 &other_level, &other_trust ) ) {
      /* lpanic( "LOMAC(lomac_monitor.c) - " *
       * "monitor_kill got dest PLM miss" ); */
      /* Assume the user process specified a non-existent pid */
      return( -EPERM );
    }
    if( current_level < other_level ) {

#ifdef VERBOSELOG
      procname_s = lprocess_getname( p_current );
      printk( KERN_INFO "process %u (%s) group %u "
	      "denied signal %u to processes %u.\n",
	      lprocess_pid( p_current ), procname_s, 
	      current_pgid, sig, pid );
      lprocess_putname( procname_s );
#endif

      return( -EPERM );
    }


  } /* if/else/if/else... pid */

  /* The pid == 0 case winds up here, which is signal the current *
   * process's pgrp, which we always allow.                       */

  return( 0 );                 /* OK to proceed with sys_kill() */

} /* monitor_kill() */



/* monitor_write()
 *
 * in:     p_inode - inode being written
 *         label_s - string containing name of syscall used for write
 * out:    nothing
 * return: value    condition
 *         -----    ---------
 *           0      allow write to proceed
 *           1      disallow write
 *
 * This function determines the level of the object indicated by
 * `p_inode'.  In order to accomplish this feat, it must first
 * determine what kind of object `p_inode' refers to, and then query
 * the level mapping appropriate to that kind of object.
 *
 *
 */

int monitor_write( lvnode_t *p_inode, char *label_s ) {

  level_t inode_level = LOMAC_INVALID_LEVEL;  /* level of `p_inode' */
  level_t proc_level  = LOMAC_INVALID_LEVEL;  /* level of current process */
  trust_t proc_trust  = e_invalid;            /* current proc trust setting */
  lprocess_t *p_current;                      /* the current process */
  lpid_t current_pid;                         /* pid of the current process */

#ifdef VERBOSELOG
  lvnode_id_t lvnode_id;    /* unique id for `p_inode' */
  char *procname_s = NULL;  /* holds name of current exe based on argv[ 0 ] */
#endif

  p_current = lprocess_current();
  current_pid = lprocess_pgrp( p_current );

#ifdef PARANOID
  if( !p_inode ) {
    lpanic( "LOMAC(monitor_write) - null inode pointer" );
  }
  if( !label_s ) {
    lpanic( "LOMAC(monitor_write) - null label string" );
  }
#endif

  /* STEP 0:  Determine the level of the current process.  If the level
   *          is LOMAC_HIGHEST_LEVEL, return immediately, since processes
   *          at the highest level can write to any object.
   */

  if( plm_get( current_pid, &proc_level, &proc_trust ) ) {
    lpanic( "LOMAC(lomac_monitor.c) - monitor_write failed to get PLM entry" );
  }

  if( proc_level == LOMAC_HIGHEST_LEVEL ) {
    return( 0 );
  }


  /*
   *
   * STEP 1:  We allow writes to certain exempt objects regardless
   *          of level.  If `p_inode' is exempt, return immediately.
   *          Exempt objects include non-local domain sockets, serial
   *          devices and pseudoterminals.
   *
   */

  if( ( p_inode->i_mode & S_IFCHR ) &&        /* if character device */
      ( ( MAJOR( p_inode->i_rdev ) == 5 ) ||     /* cua  */
	( MAJOR( p_inode->i_rdev ) == 4 ) ||     /* tty  */
	( MAJOR( p_inode->i_rdev ) == 3 ) ||     /* ttyp */
	( MAJOR( p_inode->i_rdev ) == 2 ) ) ) {  /* pty  */
    
    return( 0 );
  }

  if( lvnode_is_nonlocal_socket( p_inode ) ) {
    return( 0 );
  }
 

  /* 
   *
   * STEP 2:  Unnamed pipes are also a special case.  We always allow
   *          writes to these pipes, but we demote the pipe to the
   *          level of the writing process.
   *
   */

  if( lvnode_is_pipe( p_inode ) ) {
    fsolm_set_for_new_object( p_inode, 0 );
    return( 0 );
  }



  /*
   *
   * STEP 3:  Determine the level of `p_inode' according to the FSOLM.
   *
   */

  inode_level = fsolm_get( p_inode );



  /*
   *
   * STEP 4:  Compare the level of `p_inode' to the level of the
   *          current process, and disallow the write if necessary.
   *
   */

  if( inode_level > proc_level ) {

#ifdef VERBOSELOG

    /* We're going to disallow the write, so we must construct      *
     * a meaningful log message.  Make `procname_s' point to a      *
     * kernel-space copy of the `current' process's `argv[0]'.      *
     * argv[0] should contain the name of the program the `current' *
     * process is running - we'll use it in the log message.        */

    procname_s = lprocess_getname( p_current );
    lvnode_to_id( p_inode, &lvnode_id );
    printk( KERN_INFO "process %u (%s) at level %u denied %s to level %u"
	              " inode number %lu device %u\n",
	    current_pid, procname_s,
	    proc_level,
	    label_s, inode_level,
	    lvnode_id.lvnode_number, 
	    lvnode_id.device_number );
    lprocess_putname( procname_s );

#endif

    /* deny the write */
    return( 1 );

  } /* if denial required */

  return( 0 );

} /* monitor_write() */



/* has_sufficient_privilege()
 *
 * in:     nothing
 * out:    nothing
 * return: value    condition
 *         -----    ---------
 *           1      allow privileged system call
 *           0      disallow privileged system call
 * 
 * Some important system calls can be called only by processes with
 * sufficient privilege.  Currently, it is sufficient simply to be at
 * LOMAC_HIGHEST_LEVEL.  Things may get more complex later.
 *
 */

int has_sufficient_privilege( void ) {

  level_t level;  /* level of current process */
  trust_t trust;  /* trust setting of current process */

  if( plm_get( lprocess_pgrp( lprocess_current() ), &level, &trust ) ) {
    lpanic( "LOMAC(lomac_monitor.c) - can't get pid level for priv check" );
  }


  if( level != LOMAC_HIGHEST_LEVEL ) {
#ifdef VERBOSELOG
    printk( KERN_INFO
	    "process %u pgrp %u denied privileged syscall\n", 
	    lprocess_pid( lprocess_current() ),
	    lprocess_pgrp( lprocess_current() ) );
#endif
    return( 0 );
  } else {
    return( 1 );
  }

} /* has_sufficient_privilege() */




/* handle_unix_domain_bind()
 *
 * in:     p_socket_inode - inode representing bound UNIX-domain socket
 *         p_name_inode   - inode representing UNIX-domain name
 * out:    nothing
 * return: nothing
 *
 * When a process binds a socket to a unix (local) domain name, the name
 * is represented by its own inode.  This inode is created on the first
 * bind to the name.  We must respond to the creation/opening of the
 * name inode by updating the VLM, in the same way we do for the
 * sys_open() and sys_creat() system calls.
 *
 * When the process reads/writes from the socket, it reads/writes from
 * the inode representing the socket itself, not the inode that
 * represents the unix-domain name.  Thus, without our intervention,
 * the level of the name's inode would have no effect on the reading
 * and writing behavior of the process.  This is an unacceptable
 * situation, because the LOMAC policy input specifies levels for
 * unix-domain names with the expectation that those levels will
 * govern the behavior of all sockets bound to the names.  To provide
 * this desired behavior, on bind(), we must set the level of the
 * socket inode to be the same as the level of the unix-domain name
 * inode.
 *
 */

void handle_unix_domain_bind( lvnode_t *p_socket_inode,
			      lvnode_t *p_name_inode ) {

  /* Step 1 - Update the VLM to reflect the opening/creation of *
   *          the inode representing the unix-domain name.      */

  /* Note - we're treating bind() linke sys_creat() here - even if the *
   * name existed before, we're assuming that the bind() eliminates    *
   * its previous contents.  Is this OK?  Um....                       */
  fsolm_set_for_new_object( p_name_inode, 1 );


  /* Step 2 - Ensure that the level of the socket `p_socket_inode', *
   *          which is bound to the unix-domain name represented by *
   *          `p_name_inode', is the same as the level of           *
   *          `p_name_inode'.                                       */

  fsolm_set_to_same( p_name_inode, p_socket_inode );

  /* printk( "bound socket %lu, %u to name %lu, %u\n",
   * 	    p_socket_inode->i_ino, p_socket_inode->i_dev,
   *	    p_name_inode->i_ino, p_name_inode->i_dev );
   */
  
} /* handle_unix_domain_bind() */


/* handle_unix_domain_accept_connect()
 *
 * in:     p_old_inode - the inode whose level is already correctly set.
 *                       For accept, this is the inode representing the
 *                       original socket used to receive the incoming
 *                       connection.  For connect, this is the inode
 *                       representing the socket created at the server
 *                       by sys_accept().
 *         p_new_inode - the inode whose level may need to be set.  For
 *                       accept, this is the inode representing the socket
 *                       created for the connection by the sys_accept()
 *                       call.  For connect, this is the inode representing
 *                       the client's socket.
 * out:    nothing
 * return: nothing
 *
 * When a server calls sys_accept(), the kernel creates a new socket
 * based on the original one used to receive the connection request.
 * We must ensure that the new socket has the same level as the
 * original one.
 *
 * When a client calls sys_connect(), the kernel connects its socket
 * with the socket the server created with sys_accept().  We must
 * ensure that the client's socket has the same level as the server's
 * socket.
 *
 */

void handle_unix_domain_accept_connect( lvnode_t *p_old_inode,
					lvnode_t *p_new_inode ) {

  fsolm_set_to_same( p_old_inode, p_new_inode );

} /* handle_unix_domain_accept_connect() */



/* monitor_pgrp_change()
 *
 * in:     p_changing_process - process whose pgrp we want to change
 *         old_pgrp           - current pgrp
 *         new_pgrp           - pgrp we want to change to, may be
 *                              invalid pgrp number
 * out:    nothing
 * return: value     condition
 *         -----     ---------
 *           0       Allow pgrp change.
 *           1       Disallow pgrp change.
 *
 *
 * If we can't find a PLM mapping for `new_pgrp', we assume the user-space
 * process specified a non-existant pgrp number and indicate disallow.
 *
 */

int monitor_pgrp_change( lprocess_t *p_changing_process, lpid_t old_pgrp,
			 lpid_t new_pgrp ) {

  level_t old_level, new_level;  /* levels of `old,new_pgrp' */
  trust_t old_trust, new_trust;  /* trust settings of `old,new_pgrp' */

  /* We'll always allow a process to change to its own new pgrp. */
  if( lprocess_pid( p_changing_process ) == new_pgrp ) {
    return( 0 );
  }

  if( plm_get( old_pgrp, &old_level, &old_trust ) ) {
    lpanic( "LOMAC(lomac_monitor.c) - source pgrp PLM miss on change" );
  }

  if( plm_get( new_pgrp, &new_level, &new_trust ) ) {
    return( -EPERM );  /* assume we're OK and `new_pgrp' doesn't exist */
  }

  if( old_level < new_level ) {
    printk( KERN_INFO 
	    "denied pid %d change from [group,level] [%d,%d] to [%d,%d] "
	    "by pid %d\n",
	    lprocess_pid( p_changing_process ),
	    old_pgrp, old_level, new_pgrp, new_level,
	    lprocess_pid( lprocess_current() ) );
    return( 1 );
  } else {
    /* printk( KERN_INFO 
     *	    "allowed pid %d change from [group,level] [%d,%d] to [%d,%d] "
     *	    "by pid %d\n",
     *      lprocess_pid( p_changing_process ),
     *	    old_pgrp, old_level, new_pgrp, new_level,
     *	    lprocess_pid( lprocess_current() ) );
     */
    return( 0 );
  }

} /* monitor_pgrp_change() */



/* handle_exec()
 *
 * in:     filename_s - absolute path to binary being exec'd
 * out:    nothing
 * return: nothing
 *
 * Sets the trusted attribute of a process's PLM entry appropriately
 * when it execs a new binary image.
 *
 * Also treats the exec() as a "read" of the binary image `filename_s'.
 *
 */

void handle_exec( const char *filename_s ) {

  lpid_t current_pid;       /* pid of current process */
  level_t current_level;    /* level of current process */
  trust_t current_trust;    /* trust setting of current process */
 
  lvnode_t *p_lvnode;       /* lvnode corresponding to `filename_s' */

  /* Completion of the exec system call is effectively a "read" of the *
   * binary image indicated by `filename_s'.  Determine which lvnode   *
   * is named y `filename_s' and do read mediation on it.              */
  if( ( p_lvnode = lvnode_from_path( filename_s ) ) ) {
    monitor_read( p_lvnode, "exec" );
  }


  current_pid = lprocess_pgrp( lprocess_current() );

  /* We'll allow trusted programs only at LOMAC_HIGHEST_LEVEL for now. */
  if( plm_get( current_pid, &current_level, &current_trust ) ) {
    lprint( "pgrp %d caused\n", current_pid );
    lpanic( "LOMAC(lomac_tpl.c) - handle_exec caught PLM miss" );
  }
  if( ( is_trusted_program( filename_s ) ) &&
      ( LOMAC_HIGHEST_LEVEL == current_level ) ) {

    plm_set( current_pid, current_level, e_neverdemote );

#ifdef VERBOSELOG
    printk( KERN_INFO "pid %u executes trusted program %s\n",
	    current_pid, filename_s );
#endif

  } else {
    /* else clear trust, in case we are the child of a trusted program */
    plm_set( current_pid, current_level, e_untrusted );
  }

} /* handle_exec() */

