/*************************************************************************
 *
 * lomac_nlm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 *
 * References:
 *
 * Alan Cox `Network Buffers and Memory Management' in Linux Journal #30,
 * October, 1996.
 *
 * McKusick et. al. `The Design and Implementation of the 4.4 BSD
 * Operating System', Addison-Wesley, 1996, chapters 11 and 12.
 *
 * David A. Rusling `The Linux Kernel' section 8.6: Network Devices, March
 * 1998. (A description of the linux kernel, available on the web.)
 *
 *************************************************************************/

#include "kernel_kernel.h"
#include "kernel_lnic.h"

#include "lomac_nlm.h"

#ifdef PARANOID
#define NLM_HASHNODE_MAGIC   0x2Facaded
#endif
#define NLM_HASHNUM          0x07
#define NLM_HASH( p_nic )      ( ((unsigned int)(p_nic)) & NLM_HASHNUM )

typedef struct nlm_hashnode {
#ifdef PARANOID
  unsigned int magic;           /* magic number for paranoid assertions */
#endif
  lnic_t *p_nic;                /* pointer to `struct device' for NIC */
  level_t level;                /* level equivalent to group id `gid' */
  struct nlm_hashnode *p_next;  /* next bucket in hash chain */
} nlm_hashnode_t;

nlm_hashnode_t *nlm_hash_table[ NLM_HASHNUM + 1 ];


int     set_nic_level( char *, level_t );




/* ------------------------ Public Functions -------------------------- */

/* init_nlm()
 *
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 */

void init_nlm() {

  /* before we begin, clear `nlm_hash_table' */
  memset( (void *)nlm_hash_table, 0,
	  ( sizeof( nlm_hashnode_t *) * NLM_HASHNUM + 1 ) );

#ifdef SERVER_POLICY
  set_nic_level( "eth0", LOMAC_LOWEST_LEVEL );
  set_nic_level( "eth1", 2 );
#endif

}

/* nic_to_level()
 *
 * in:     p_nic - pointer to `struct device' representing NIC
 * out:    nothing
 * return: level of NIC indicated by `p_nic'
 *
 * Searches the NLM for an entry indicating the level of the NIC 
 * indicated by `p_nic'.  If an entry is found, returns the level.
 * Otherwise, returns LOMAC_LOWEST_LEVEL.  That is, the default
 * level for NICs is LOMAC_LOWEST_LEVEL unless specifically stated
 * otherwise in the policy input.
 *
 */

level_t nic_to_level( lnic_t *p_nic ) {

  nlm_hashnode_t *p_hashnode;  /* points to hashnodes in the NLM hash table */

#ifdef PARANOID
  if( !p_nic ) {
    lpanic( "LOMAC(lomac_nlm.c) null NIC pointer" );
  }
#endif


  p_hashnode = nlm_hash_table[ NLM_HASH( p_nic ) ];
  while( p_hashnode ) {

#ifdef PARANOID
    if( p_hashnode->magic != NLM_HASHNODE_MAGIC ) {
      lpanic( "LOMAC(lomac_nlm.c) "
	     "Bad magic number on NLM hashnode query" );
    }
    if( ( p_hashnode->level < LOMAC_LOWEST_LEVEL ) ||
	( p_hashnode->level > LOMAC_HIGHEST_LEVEL ) ) {
      lpanic( "LOMAC(lomac_nlm.c) Out of bounds level in NLM" );
    }
#endif
      
    if( p_hashnode->p_nic == p_nic ) {
      return( p_hashnode->level );
    }

    p_hashnode = p_hashnode->p_next;

  } /* for all hashnodes in chain */

  /* If we reach here, then `p_nic' does not exist in the NLM. *
   * Return the default NIC level of LOMAC_LOWEST_LEVEL.       */
  return( LOMAC_LOWEST_LEVEL );

} /* nic_to_level() */



/* ------------------------ Private Functions -------------------------- */


/* set_nic_level()
 *
 * in:     name_s - name of NIC
 *         level  - level for NIC named by `name_s'
 * out:    nothing
 * return: value       condition
 *         -------
 *         0           success
 *         1           no NIC with name `name_s' found
 *
 * The Linux kernel represents NICs with instances of the `struct
 * netdevice' structure.  Each `struct netdevice' has a `name' field
 * which points to a string containing the human-readable name of the
 * NIC, such as "eth0".  To make life easier for administrators, LOMAC
 * expects its policy input to indentify NICs by their human-readable
 * name.  However, once inside the LKM, we want to avoid identifying
 * NICs by their human-readble names, because this form of
 * identification would require lots of slow string-comparisons.  So,
 * the nic_to_level() function takes human-readable names as input (to
 * make admins happy), and converts them to the equivalent `struct
 * netdevice' pointers before stuffing them into the NLM (to make
 * LOMAC happy).
 *
 */

int set_nic_level( char *name_s, level_t level ) {

  lnic_t *p_nic = NULL;        /* points to the device struct for `name_s' */
  nlm_hashnode_t *p_hashnode;  /* points to hashnodes in the NLM hash table */
  int hash_chain_index;        /* holds results of NLM_HASH() */

#ifdef PARANOID
  if( !name_s ) {
    lpanic( "LOMAC(lomac_nlm.c) NULL name string pointer" );
  }
  if( ( level < LOMAC_LOWEST_LEVEL ) || ( level > LOMAC_HIGHEST_LEVEL ) ) {
    lpanic( "LOMAC(lomac_nlm.c) level out of bounds" );
  }
#endif

  /*
   *
   * STEP 1: Convert the `name_s' string to the corresponding `struct
   *         device' pointer and put it in `p_nic'.  We know that all
   *         NICs are represented by a `struct device', and all
   *         `struct device' instances are hanging in a linked list
   *         off of `dev_base'.  Iterate through the `dev_base' list,
   *         looking for a `struct device' with name `name_s'.  If
   *         found, put its address in `p_nic'.  Otherwise, return 1.
   */

  for( p_nic = lnic_first_device(); p_nic != NULL;
       p_nic = lnic_next( p_nic ) ) {
    if( !strcmp( lnic_name( p_nic ), name_s ) ) {
#ifdef VERBOSELOG
      lprint( KERN_INFO "set NIC %s to level %d\n", 
	      lnic_name( p_nic ), level );
#endif
      break;
    }
  }
  if( !p_nic ) {
#ifdef VERBOSELOG
    lprint( KERN_INFO "NIC %s not found\n", name_s );
#endif
    return( 1 );   /* there was no device with name `name_s' */
  }


  /*
   *
   * STEP 2: Add an entry for `p_nic' and `level' into the NLM
   * 
   */

  /* First, try to find an existing entry for `p_nic' in *
   * the NLM hash table and update its level field.      */
  
  hash_chain_index = NLM_HASH( p_nic );
  p_hashnode = nlm_hash_table[ hash_chain_index ];
  while( p_hashnode ) {

#ifdef PARANOID
    if( p_hashnode->magic != NLM_HASHNODE_MAGIC ) {
      lpanic( "Lomac (lomac_nlm.c) : Bad magic number on NLM hashnode set" );
    }
#endif
      
    if( p_hashnode->p_nic == p_nic ) {
      p_hashnode->level = level;   /* a node already exists - update it. */

#ifdef PARANOID
      if( level != nic_to_level( p_nic ) ) {
	lpanic( "LOMAC(lomac_nlm.c) failed to set NIC level" );
      }
#endif

      return( 0 );
    }

    p_hashnode = p_hashnode->p_next;

  } /* for all hashnodes in chain */

  /* If we reach here, then an entry for `p_nic' does not exist in *
   * the NLM.  we must add a new hashnode to the NLM for `p_nic'   *
   * at the head of the chain indicated by `hash_chain_index'.     */

  p_hashnode = (nlm_hashnode_t *)lmalloc( sizeof( nlm_hashnode_t ) );

#ifdef PARANOID
  if( !p_hashnode ) {
    lpanic( "Lomac (lomac_nlm.c) : Unable to allocate new NLM hashnode" );
  }
  p_hashnode->magic = NLM_HASHNODE_MAGIC;
#endif

  p_hashnode->p_nic = p_nic;
  p_hashnode->level = level;
  p_hashnode->p_next = nlm_hash_table[ hash_chain_index ];
  nlm_hash_table[ hash_chain_index ] = p_hashnode;

#ifdef PARANOID
  if( level != nic_to_level( p_nic ) ) {
    lpanic( "LOMAC(lomac_nlm.c) failed to set NIC level" );
  }
#endif

  return( 0 );

} /* set_nic_level() */


